import java.io.InputStream;
import java.io.OutputStream;
import java.io.IOException;

class Agent implements Runnable
{
   private static final int BUFFER_SIZE = 4096;

   private InputStream  mIn         = null;
   private OutputStream mOut        = null;
   private AgentMonitor mMonitor    = null;
   private byte         mBuffer[]   = null;
   private String       mName       = null;
   private boolean      mEcho       = false;
   private boolean      mBlock      = false;
   private boolean      mXBlock     = false;
   private int          mID         = 0;

   public Agent(
      InputStream in,
      OutputStream out,
      AgentMonitor monitor,
      String name,
      boolean echo,
      boolean block,
      boolean xBlock,
      int idNumber
   )
   {
      mIn      = in;
      mOut     = out;
      mMonitor = monitor;
      mName    = name;
      mEcho    = echo;
      mBlock   = block;
      mXBlock  = xBlock;
      mID      = idNumber;

      mBuffer  = new byte[BUFFER_SIZE];
      Thread t = new Thread(this);

      t.start();
   }

   public void run()
   {
      try 
      {
         int bytesRead = 0;

         while(true)
         {
            if((bytesRead = mIn.read(mBuffer, 0, BUFFER_SIZE)) == -1)
               break;

            if(mEcho)
               doEcho(mBuffer, bytesRead);

            mOut.write(mBuffer, 0, bytesRead);
         }
      }
      catch(IOException e) {}

      mMonitor.agentHasDied(this);
   }

   private static String pad(
      int value, int width, char padChar, boolean doHex
   )
   {
      String s = "";

      if(doHex)
      {
         s = s + Integer.toHexString(value);
      }
      else
      {
         s = s + value;
      }

      int length = s.length();

      StringBuffer sb = new StringBuffer();

      int needed = (length >= width ? 0 : width - length);

      for(int i = 0; i < needed; i ++)
         sb.append(padChar);

      sb.append(s);

      return(sb.toString());
   }

   private void doEcho(byte buffer[], int nBytes)
   {
      synchronized(System.out)
      {
         String timeString = "" + System.currentTimeMillis();
         int length        = timeString.length();

         timeString = timeString.substring(0, length - 3)
                    + "." + timeString.substring(length - 4);

         System.out.println(
            "[" + mID + ": " + mName + ", (" + timeString + "), " +
            "(" + nBytes + " bytes)]"
         );

         StringBuffer sb = new StringBuffer("");

         if(mBlock || mXBlock)
         {
            for(int i = 0; i < nBytes; i += 10)
            {
               sb.append(pad(i, 8, '0', false));

               StringBuffer display = new StringBuffer("|");

               for(int j = 0; j < 10; j ++)
               {
                  sb.append(j == 0 ? '|' : ' ');

                  if(i + j < nBytes)
                  {
                     int value = (mBuffer[i + j] & 0xFF);

                     sb.append(pad(value, 3, ' ', mXBlock));

                     char c = (char)value;

                     if(value >= ' ' && value <= '~')
                     {
                        display.append((char)value);
                     }
                     else
                     {
                        display.append(".");
                     }
                  }
                  else
                  {
                     sb.append("   ");
                  }
               }

               display.append("|");

               sb.append(display).append("\n");
            }
         }
         else
         {
            for(int i = 0; i < nBytes; i ++)
            {
               boolean useChar = false;
               int value = (mBuffer[i] & 0xFF);

               if(value >= ' ' && value <= '~')
               {
                  useChar = true;
               }
               else if(value == '\n' || value == '\r')
               {
                  useChar = true;
               }

               if(!useChar)
                  sb.append(".");
               else
                  sb.append((char)value);
            }
         }

         System.out.println(sb.toString());
         System.out.flush();
      }
   }
}
